package env;

import node.*;
import node.declNode.*;
import type.*;
import type.basicType.*;
import code.*;
import text.*;

public class Decl {

	public final static int DECL_NONE		= 0;
	public final static int DECL_CLASS		= 1;
	public final static int DECL_METHOD		= 2;
	public final static int DECL_VAR		= 3;
	public final static int DECL_PARAM		= 4;

	public static String sectionText( int section ) {
		switch ( section ) {
			case DECL_NONE:
				return "none";
			case DECL_CLASS:
				return "class";
			case DECL_METHOD:
				return "method";
			case DECL_VAR:
				return "var";
			case DECL_PARAM:
				return "param";
			default:
				return "section_" + section;
			}
		}

	private String ident;
	private int section;
	private Type type;
	private DeclNode declNode;
	private Env env;
	private Offset offset;
	private Decl overRiddenDecl;

	public String ident() { return ident; }
	public int section() { return section; }
	public Type type() { return type; }
	public DeclNode declNode() { return declNode; }
	public Env env() { return env; }
	public Offset offset() { return offset; }
	public Decl overRiddenDecl() { return overRiddenDecl; }

	public Decl( 
		String ident, 
		int section, 
		DeclNode declNode, 
		Env env ) {
		Print.error().debugln( "Decl( " + ident + ", " 
			+ sectionText( section ) + ", ... )" );
		this.ident = ident;
		this.section = section;
		this.declNode = declNode;
		this.env = env;
		}

	public boolean equals( Decl otherDecl ) {
		return ident.equals( otherDecl.ident() ) 
			&& type.equals( otherDecl.type() );
		}

	public void setType( Type type ) {
		Print.error().debugln( "Type " + ident + " = " + type );
		this.type = type;
		}
		
	public void genOffset() {
		switch ( section ) {
			case DECL_METHOD:
				if ( env.envKind() == Env.ENV_CLASS ) {
					Env extendedEnv = env.extendedEnv();
					if ( extendedEnv != null )
						overRiddenDecl = 
							extendedEnv.searchExtended( ident );
					if ( overRiddenDecl != null ) {
						if ( overRiddenDecl.section() != DECL_METHOD )
							throw new Error( 
								"Overridden declaration " + overRiddenDecl 
								+ " must be method" );
						MethodType methodType = ( MethodType ) type;
						MethodType overRiddenMethodType = 
							( MethodType ) overRiddenDecl.type();
						if ( ! methodType.equals( overRiddenMethodType ) )
							throw new Error( 
								"Overridden declaration " + overRiddenDecl 
								+ " must have same type" );
						offset = overRiddenDecl.offset();
						}
					else {
						offset = env.methodOffset().inc( type.size() );
						}
					env.methodOffset().align();
					}
				else
					offset = new Offset( 0, 0 );
				break;
			case DECL_NONE:
			case DECL_CLASS:
				offset = new Offset( 0, 0 );
				break;
			case DECL_VAR:
			case DECL_PARAM:
				offset = env.varOffset().inc( type.size() );
				env.varOffset().align();
				break;
			default:
				throw new Error( 
					"Invalid section of " + sectionText( section() ) );
			}
		Print.error().debugln( "Offset " + ident + " = " + offset );
		}

	public String declName() { /* Simple name as declared */
//		Print.error().debugln( "declName " + ident + " = " + offset );
		switch ( section ) {
			case DECL_NONE:
				return ident;
			case DECL_CLASS:
			case DECL_METHOD:
			case DECL_VAR:
			case DECL_PARAM:
				return ident + "_" + Integer.toHexString( offset.compile() );
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public String useName() {
		switch ( section ) {
			case DECL_NONE:
			case DECL_CLASS:
				return declName();
			case DECL_VAR:
			case DECL_PARAM:
				switch ( env.envKind() ) {
					case Env.ENV_GLOBAL:
					case Env.ENV_METHOD:
						return declName();
					case Env.ENV_CLASS:
						return "field." + declName();
					}
			case DECL_METHOD:
				switch ( env.envKind() ) {
					case Env.ENV_GLOBAL:
						return "methodCode." + declName();
					case Env.ENV_CLASS:
						return "method." + declName();
					case Env.ENV_METHOD:
					}
			default:
				throw new Error( 
					"Invalid section of " + sectionText( section() ) );
			}
		}

	public String absoluteName() {
		return env.absoluteName() + "." + useName();
		}

	public void allocateSpace() {
		Code.publicLabelDefn( declName() );
		Code.space( type.size().compile() );
		Code.align();
		}

	public Displacement address() {
		switch ( env().envKind() ) {
			case Env.ENV_GLOBAL:
				return new Indirect( new Literal( absoluteName() ) );
			case Env.ENV_CLASS:
				//	Only works for fields
				return new Displacement( absoluteName(), SpecialReg.instPtr );
			case Env.ENV_METHOD:
				return new Displacement( declName(), SpecialReg.framePtr );
			default:
				throw new Error( 
					"Invalid section of " + sectionText( section() ) );
			}
		}

	public void store( Register source ) {
		switch ( env().envKind() ) {
			case Env.ENV_GLOBAL:
				Code.loadImm( IntType.type, SpecialReg.assemTemp, absoluteName() );
				Code.store( type, source, new Indirect( SpecialReg.assemTemp ) );
				break;
			case Env.ENV_METHOD:
				Code.store( type, source, 
					new Displacement( useName(), SpecialReg.framePtr ) );
				break;
			case Env.ENV_CLASS:
				Code.store( type, source, 
					new Displacement( useName(), SpecialReg.instPtr ) );
				break;
			default:
				throw new Error( 
					"Invalid section of " + sectionText( section() ) );
			}
		}

	public String toString() {
		switch ( section ) {
			case DECL_NONE:
			case DECL_METHOD:
			case DECL_VAR:
			case DECL_PARAM:
				return type + " " + ident + ";";
			case DECL_CLASS:
				return "class " + ident + ";";
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}

		}
	}
