package env;

import node.*;
import node.declNode.*;
import type.*;
import runEnv.*;
import code.*;
import text.*;

public class Env {

	public final static int ENV_NONE		= 0;
	public final static int ENV_CLASS		= 1;
	public final static int ENV_GLOBAL		= 2;
	public final static int ENV_METHOD		= 3;

	private int envKind;
	private Decl decl;
	private DeclList localDecls = new DeclList();
	private Env enclosingEnv;
	private Env extendedEnv;
	private Offset varOffset = new Offset();
	private Offset methodOffset = new Offset();
	private Offset paramStart = new Offset();
	private DeclList methodTable = new DeclList();
	private long tempUsage;

	public int envKind() { return envKind; }
	public Env enclosingEnv() { return enclosingEnv; }
	public Env extendedEnv() { return extendedEnv; }
	public Offset varOffset() { return varOffset; }
	public Offset methodOffset() { return methodOffset; }
	public Offset paramStart() { return paramStart; }
	public Decl decl() { return decl; }
	public DeclList localDecls() { return localDecls; }
	public DeclList methodTable() { return methodTable; }
	
	public static String envKindText( int envKind ) {
		switch ( envKind ) {
			case ENV_NONE:
				return "none";
			case ENV_CLASS:
				return "class";
			case ENV_GLOBAL:
				return "global";
			case ENV_METHOD:
				return "method";
			default:
				return "envKind_" + envKind;
			}
		}

	public void addTempUsage( long tempUsage ) {
		this.tempUsage |= tempUsage;
		}
	
	public long tempUsage() {
		return tempUsage;
		}

	public Env( int envKind, Env enclosingEnv, Env extendedEnv ) {
		this.envKind = envKind;
		this.enclosingEnv = enclosingEnv;
		this.extendedEnv = extendedEnv;
		}
	
	public Env( int envKind, Env enclosingEnv ) {
		this( envKind, enclosingEnv, null );
		}
	
	public Env( int envKind ) {
		this( envKind, null, null );
		}
	
	public Decl methodDecl() {
		//	Used for return statements
		if ( envKind == ENV_METHOD )
			return decl;
		else if ( enclosingEnv == null )
			return null; 
		else
			return enclosingEnv.methodDecl();
		}

	public Decl classDecl() {
		//	used for "this".
		if ( envKind == ENV_CLASS )
			return decl;
		else if ( enclosingEnv == null )
			return null; 
		else
			return enclosingEnv.classDecl();
		}
	
	public boolean isExtensionOf( Env env ) {
		if ( env == null )
			return false;
		else if ( this == env )
			return true;
		else
			return isExtensionOf( env.extendedEnv );
		}
/*
These methods to search for an identifier 
*/
	public Decl searchLocal( String ident ) {
		return localDecls.searchDeclList( ident );
		}
		
	public Decl searchExtended( String ident ) {
		Decl localDecl;
		localDecl = searchLocal( ident );
		if ( localDecl != null )
			return localDecl;
		else if ( extendedEnv == null )
			return null;
		else
			return extendedEnv.searchExtended( ident );
		}

	public Decl searchEnv( String ident ) {
		Decl localDecl;
		localDecl = searchLocal( ident );
		if ( localDecl != null )
			return localDecl;
		Decl extendedDecl;
		if ( extendedEnv == null )
			extendedDecl = null;
		else
			extendedDecl = extendedEnv.searchExtended( ident );
		Decl enclosingDecl;
		if ( enclosingEnv == null )
			enclosingDecl = null;
		else
			enclosingDecl = enclosingEnv.searchEnv( ident );
		if ( extendedDecl != null && enclosingDecl != null )
			throw new Error( "Ambiguous declaration of " + ident );
		if ( extendedDecl != null && enclosingDecl == null )
			return extendedDecl;
		else if ( enclosingDecl != null && extendedDecl == null )
			return enclosingDecl;
		else
			return null;
		}

	public Decl declare( 
		String ident, 
		int section, 
		DeclNode declNode ) {
		if ( searchLocal( ident ) != null )
			throw new Error( "Identifier " + ident + " redeclared in same environment" );
		return localDecls.declare( ident, section, declNode, this );
		}
	
	public void setDecl( Decl decl ) {
		this.decl = decl;
		}

	public String declName() {
		return decl.declName();
		}

	public String useName() {
		return decl.useName();
		}

	public String absoluteName() {
		if ( enclosingEnv == null )
			return declName();
		else
			return enclosingEnv.absoluteName() + "." + useName();
		}
		
	public String toString() {
		return absoluteName();
		}
	
	public DeclList paramDecls() {
		return localDecls.paramDecls();
		}

	public void setClassOffset() {
		if ( extendedEnv == null ) {
			varOffset.set( 1, Type.WORD );
			methodOffset().set( 0, 0 );
			}
		else {
			varOffset.set( extendedEnv.varOffset() );
			methodOffset.set( extendedEnv.methodOffset() );
			}
		}

	public void createMethodTable( DeclList methodTable ) {
		if ( extendedEnv != null )
			extendedEnv.createMethodTable( methodTable );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_METHOD:
					methodTable.setElementAt( decl, decl.offset().interp() );
					break;
				}
			}
		}

	public void createMethodTable() {
		createMethodTable( methodTable );
		}
		
	public void initInstance( RunEnv runEnv ) throws ReturnException {
		if ( extendedEnv != null )
			extendedEnv.initInstance( runEnv );
		ClassDeclNode classDeclNode = ( ClassDeclNode ) decl().declNode();
		DeclStmtListNode body = classDeclNode.body();
		body.eval( runEnv );
		}

	public void genTempSpace() {
		Code.enter( "local" );
		int maxTemp = Usage.maxTemp( tempUsage );
		for ( int i = 0; i < maxTemp; i++ ) {
			Code.publicLabelDefn( "sav" + i );
			Code.space( Type.WORD );
			}
		Code.publicLabelDefn( "max" );
		Code.exit( "local" );
		}
		
	public void genLocalDecls() {
		Code.enter( "local" );
		int maxTemp = Usage.maxTemp( tempUsage );
		for ( int i = 0; i < maxTemp; i++ ) {
			Code.publicLabelDefn( "sav" + i );
			Code.space( Type.WORD );
			}
		
		Code.publicLabelDefn( "locals" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_VAR:
					decl.declNode().genDeclCode();
				}
			}
		Code.publicLabelDefn( "params" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_PARAM:
					decl.declNode().genDeclCode();
				}
			}
		Code.publicLabelDefn( "max" );
		Code.exit( "local" );
		}
		
	public void genFieldDeclOffsetCode() {
		Code.comment( "Field table layout for " + this );
		if ( extendedEnv == null ) {
			Code.enterBlock( "field", null, null );
			Code.enter( "local" );
			Code.publicLabelDefn( "methodTablePtr" );
			Code.space( Type.WORD );
			}
		else {
			Code.enterBlock( "field", extendedEnv.absoluteName() 
				+ ".field", extendedEnv.absoluteName() + ".field" );
			Code.enter( "local" );
			}
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_VAR:
					decl.declNode().genDeclCode();
					break;
				}
			}
		Code.publicLabelDefn( "max" );
		Code.exit( "local" );
		Code.exitBlock( "field" );
		}

	public void genMethodDeclOffsetCode() {
		Code.comment( "Method table layout for " + this );
		if ( extendedEnv == null )
			Code.enterBlock( "method", null, null );
		else
			Code.enterBlock( "method", 
				extendedEnv.absoluteName() + ".method", 
				extendedEnv.absoluteName() + ".method" );
		Code.enter( "local" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_METHOD:
					if ( decl.overRiddenDecl() == null )
						decl.allocateSpace();
					break;
				}
			}
		Code.publicLabelDefn( "max" );
		Code.exit( "local" );
		Code.exitBlock( "method" );
		}
		
	public void genGlobalDeclCode() {
		Code.comment( "Global variable declarations for " + this );
		Code.enter( "data" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_VAR:
					decl.declNode().genDeclCode();
					break;
				}
			}
		Code.exit( "data" );
		}

	public void genMethodDeclCode() {
		Code.comment( "Code for methods for " + this );
		Code.enterBlock( "methodCode", null, null );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_METHOD:
					decl.declNode().genDeclCode();
					break;
				}
			}
		Code.exitBlock( "methodCode" );
		}
		
	public void genMethodTable() {
		Code.comment( "Actual method table for " + this );
		Code.enterBlock( "methodTable", null, null );
		Code.enter( "const" );
		Code.publicLabelDefn( "enter" );
		for ( int i = 0; i < methodTable.size(); i++ ) {
			Decl decl = methodTable.elementAt( i );
			Code.quad( decl.env().absoluteName() 
				+ ".methodCode." + decl.declName() + ".enter" );
			}
		Code.exit( "const" );
		Code.exitBlock( "methodTable" );
		}
	
	public void genClassDeclCode() {
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.DECL_CLASS:
					decl.declNode().genDeclCode();
					break;
				}
			}
		}
	}
