package env;

import node.*;
import node.declNode.*;
import type.*;
import template.*;
import text.*;

public class Decl {

	public final static int SECT_NONE		= 0;
	public final static int SECT_CLASS		= 1;
	public final static int SECT_GLOBAL		= 2;
	public final static int SECT_STATIC		= 3;
	public final static int SECT_INSTANCE	= 4;
	public final static int SECT_VALUEPARAM	= 5;
	public final static int SECT_VARPARAM	= 6;
	public final static int SECT_LOCAL		= 7;

	public final static int NOACCESS	= 0;
	public final static int PRIVATE		= 1;
	public final static int PROTECTED	= 2;
	public final static int PUBLIC		= 3;

	public static String sectionText( int section ) {
		switch ( section ) {
			case SECT_CLASS:
				return "class";
			case SECT_GLOBAL:
				return "global";
			case SECT_STATIC:
				return "static";
			case SECT_INSTANCE:
				return "instance";
			case SECT_VALUEPARAM:
				return "value param";
			case SECT_VARPARAM:
				return "var param";
			case SECT_LOCAL:
				return "local";
			default:
				return "section_" + section;
			}
		}

	public static String accessText( int accessRight ) {
		switch ( accessRight ) {
			case NOACCESS:
				return "noaccess";
			case PRIVATE:
				return "private";
			case PROTECTED:
				return "protected";
			case PUBLIC:
				return "public";
			default:
				return "accessRight_" + accessRight;
			}
		}
	
	private String ident;
	private int section;
	private int accessRight;
	private int index;
	private int offset;
	private Type type;
	private DeclNode declNode;
	private Env env;
	private Decl overRiddenDecl;

	public String ident() { return ident; }
	public int section() { return section; }
	public int accessRight() { return accessRight; }
	public int offset() { return offset; }
	public Type type() { return type; }
	public DeclNode declNode() { return declNode; }
	public Env env() { return env; }
	public Decl overRiddenDecl() { return overRiddenDecl; }

	public Decl( 
		String ident, 
		int section, 
		int accessRight, 
		Type type, 
		DeclNode declNode, 
		Env env ) {
		this.ident = ident;
		this.section = section;
		this.accessRight = accessRight;
		this.type = type;
		this.declNode = declNode;
		this.env = env;
		}

	public void setOffset( Offset offset ) {
		switch ( section ) {
			case SECT_NONE:
				break;
			case SECT_INSTANCE:
				Env extendedEnv = env.extendedEnv();
				if ( extendedEnv != null && type instanceof MethodType )
					overRiddenDecl = extendedEnv.searchExtended( 
						ident, type.template(), PROTECTED );
				if ( overRiddenDecl != null ) {
					this.offset = overRiddenDecl.offset();
					}
				else {
					this.offset = offset.inc( 1 );
					}
				break;
			case Decl.SECT_CLASS:
			case Decl.SECT_STATIC:
			case Decl.SECT_GLOBAL:
			case SECT_VARPARAM:
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				this.offset = offset.inc( 1 );
				break;
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public String declName() { /* Simple name as declared */
			return ident + "_" + offset;
		}

	public String useName() { /* Simple name as used */
		switch ( section ) {
			case SECT_NONE:
				return ident;
			case SECT_CLASS:
				return "subClass." + declName();
			case SECT_STATIC:
			case SECT_GLOBAL:
				if ( type instanceof MethodType )
					return "static.method." + declName();
				else
					return "static.field." + declName();
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					return "object.method." + declName();
				else
					return "object.field." + declName();
			case SECT_VALUEPARAM:
			case SECT_VARPARAM:
			case SECT_LOCAL:
				return declName();
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public String absoluteName() {
		return env.absoluteName() + "." + useName();
		}

	public void allocateSpace() {
		Node.publicLabelDefn( declName() );
		Node.quad();
		}

	public String address( String regName ) {
		switch ( section ) {
			case SECT_GLOBAL:
			case SECT_STATIC:
				if ( type instanceof MethodType )
					Node.loadIQ( regName, absoluteName() + ".enter" );
				else
					Node.loadIQ( regName, absoluteName() );
				return Node.indirect( regName );
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				return Node.displacement( useName(), "$fp" );
			case SECT_VARPARAM:
				Node.load( regName, Node.displacement( useName(), "$fp" ) );
				return Node.indirect( regName );
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					throw new Error( "Invalid section of " + sectionText( section() ) );
				else
					Node.load( regName, Node.displacement( "INST.fields", "$ip" ) );
				return Node.displacement( absoluteName(), regName );
			case SECT_CLASS:
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}
/*
	public void loadAddress( String regName ) {
		switch ( section ) {
			case SECT_GLOBAL:
			case SECT_STATIC:
				if ( type instanceof MethodType )
					throw new Error( "Invalid section of " + sectionText( section() ) );
				else
					Node.loadIQ( regName, absoluteName() );
				break;
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				Node.loadA( regName, Node.displacement( useName(), "$fp" ) );
				break;
			case SECT_VARPARAM:
				Node.load( regName, Node.displacement( useName(), "$fp" ) );
				break;
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					throw new Error( "Invalid section of " + sectionText( section() ) );
				else
					Node.load( "$at", Node.displacement( "INST.fields", "$ip" ) );
				Node.loadA( regName, Node.displacement( absoluteName(), "$at" ) );
				break;
			case SECT_CLASS:
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}
*/

	public void load( String regName ) {
		switch ( section ) {
			case SECT_GLOBAL:
			case SECT_STATIC:
				if ( type instanceof MethodType )
					Node.loadIQ( regName, absoluteName() + ".enter" );
				else
					Node.load( regName, absoluteName() );
				break;
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				Node.load( regName, Node.displacement( useName(), "$fp" ) );
				break;
			case SECT_VARPARAM:
				Node.load( "$at", Node.displacement( useName(), "$fp" ) );
				Node.load( regName, Node.indirect( "$at" ) );
				break;
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					Node.load( "$at", Node.displacement( "INST.methods", "$ip" ) );
				else
					Node.load( "$at", Node.displacement( "INST.fields", "$ip" ) );
				Node.load( regName, Node.displacement( absoluteName(), "$at" ) );
				break;
			case SECT_CLASS:
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public void store( String regName ) {
		switch ( section ) {
			case SECT_GLOBAL:
			case SECT_STATIC:
				Node.loadIQ( "$at", absoluteName() );
				Node.store( regName, Node.indirect( "$at" ) );
				break;
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				Node.store( regName, Node.displacement( useName(), "$fp" ) );
				break;
			case SECT_VARPARAM:
				Node.load( "$at", Node.displacement( useName(), "$fp" ) );
				Node.store( regName, Node.indirect( "$at" ) );
				break;
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					throw new Error( "Invalid section of " + sectionText( section() ) );
				else {
					Node.load( "$at", Node.displacement( "INST.fields", "$ip" ) );
					Node.store( regName, Node.displacement( absoluteName(), "$at" ) );
					}
				break;
			case SECT_CLASS:
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public void storeF( String regName ) {
		switch ( section ) {
			case SECT_GLOBAL:
			case SECT_STATIC:
				Node.loadIQ( "$at", absoluteName() );
				Node.storeF( regName, Node.indirect( "$at" ) );
				break;
			case SECT_LOCAL:
			case SECT_VALUEPARAM:
				Node.storeF( regName, Node.displacement( useName(), "$fp" ) );
				break;
			case SECT_VARPARAM:
				Node.load( "$at", Node.displacement( useName(), "$fp" ) );
				Node.storeF( regName, Node.indirect( "$at" ) );
				break;
			case SECT_INSTANCE:
				if ( type instanceof MethodType )
					throw new Error( "Invalid section of " + sectionText( section() ) );
				else {
					Node.load( "$at", Node.displacement( "INST.fields", "$ip" ) );
					Node.storeF( regName, Node.displacement( absoluteName(), "$at" ) );
					}
				break;
			case SECT_CLASS:
			default:
				throw new Error( "Invalid section of " + sectionText( section() ) );
			}
		}

	public String toString() {
		return absoluteName();
		}
	}
