package env;

import node.*;
import node.declNode.*;
import type.*;
import template.*;
import text.*;

public class Env {

	public final static int ENV_NONE		= 0;
	public final static int ENV_CLASS		= 1;
	public final static int ENV_GLOBAL		= 2;
	public final static int ENV_METHOD		= 3;

	private int envKind;
	private Env enclosingEnv;
	private Env extendedEnv;
	private Offset localOffset;
	static private Offset classOffset = new Offset( 0 );
	static private Offset staticFieldOffset = new Offset( 0 );
	static private Offset staticCodeOffset = new Offset( 0 );
	private Offset instanceFieldOffset;
	private Offset instanceCodeOffset;
	private Decl decl;
	private DeclList localDecls = new DeclList();
	private DeclList methodTable = new DeclList();

	public Env enclosingEnv() { return enclosingEnv; }
	public Env extendedEnv() { return extendedEnv; }
	public Offset localOffset() { return localOffset; }
	static public Offset classOffset() { return classOffset; }
	static public Offset staticFieldOffset() { return staticFieldOffset; }
	static public Offset staticCodeOffset() { return staticCodeOffset; }
	public Offset instanceFieldOffset() { return instanceFieldOffset; }
	public Offset instanceCodeOffset() { return instanceCodeOffset; }
	public Decl decl() { return decl; }
	public DeclList methodTable() { return methodTable; }

	public Env( int envKind, Env enclosingEnv, Env extendedEnv ) {
		this.envKind = envKind;
		this.enclosingEnv = enclosingEnv;
		this.extendedEnv = extendedEnv;
		}
	
	public Env( int envKind, Env enclosingEnv ) {
		this.envKind = envKind;
		this.enclosingEnv = enclosingEnv;
		}
	
	public Env( int envKind ) {
		this.envKind = envKind;
		}
	
	public Decl methodDecl() {
		//	Used for return statements
		if ( envKind == ENV_METHOD )
			return decl;
		else if ( enclosingEnv == null )
			return null; 
		else
			return enclosingEnv.methodDecl();
		}

	public Env classEnv() {
		//	used for "this".
		if ( envKind == ENV_CLASS )
			return this;
		else if ( enclosingEnv == null )
			return null; 
		else
			return enclosingEnv.classEnv();
		}
	
	public Decl classDecl() {
		//	used for "this".
		if ( envKind == ENV_CLASS )
			return decl;
		else if ( enclosingEnv == null )
			return null; 
		else
			return enclosingEnv.classDecl();
		}
	
	public boolean isExtension( Env env ) {
		if ( env == null )
			return false;
		else if ( this == env )
			return true;
		else
			return isExtension( env.extendedEnv );
		}
/*

These methods to search for an identifier 
that matches a given template and access right.

They do not take into account overloading.

This feature should be added.

*/
	public Decl searchLocal( String ident, Template template, int accessRight ) {
		return localDecls.searchDeclList( ident, template, accessRight );
		}
		
	public Decl searchExtended( String ident, Template template, int accessRight ) {
		Decl decl = searchLocal( ident, template, accessRight );
		if ( decl != null )
			return decl;
		else if ( extendedEnv == null )
			return null;
		else
			return extendedEnv.searchExtended( ident, template, 
				Math.max( accessRight, Decl.PROTECTED ) );
		}

	public Decl searchEnv( String ident, Template template, int accessRight ) {
		Decl localDecl;
		localDecl = searchLocal( ident, template, accessRight );
		if ( localDecl != null )
			return localDecl;
		Decl extendedDecl;
		if ( extendedEnv == null )
			extendedDecl = null;
		else
			extendedDecl = extendedEnv.searchExtended( ident, template,
				Math.max( accessRight, Decl.PROTECTED ) );
		Decl enclosingDecl;
		if ( enclosingEnv == null )
			enclosingDecl = null;
		else
			enclosingDecl = enclosingEnv.searchEnv( ident, template, accessRight );
		if ( extendedDecl != null && enclosingDecl == null )
			return extendedDecl;
		else if ( enclosingDecl != null && extendedDecl == null )
			return enclosingDecl;
		else
			return null;
		}

	public String toString() {
		return absoluteName();
		}
	
	public String declName() {
		return decl.declName();
		}

	public String useName() {
		return decl.useName();
		}

	public String absoluteName() {
		if ( enclosingEnv == null )
			return useName();
		else
			return enclosingEnv.absoluteName() + "." + useName();
		}
		
	public void setDecl( Decl decl ) {
		this.decl = decl;
		}

	public Decl declare( 
		String ident, 
		int section, 
		int accessRight, 
		Type type, 
		DeclNode declNode ) {
		return localDecls.declare( ident, section, accessRight, type, declNode, this );
		}
	
	public DeclList paramDecls() {
		DeclList paramDecls = new DeclList();
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_VARPARAM:
				case Decl.SECT_VALUEPARAM:
					paramDecls.addElement( decl );
				}
			}
		return paramDecls;
		}

	public void createMethodTable( DeclList methodTable ) {
		if ( extendedEnv != null )
			extendedEnv.createMethodTable( methodTable );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_INSTANCE:
					if ( decl.type() instanceof MethodType ) {
						methodTable.setElementAt( decl, decl.offset() );
						}
					break;
				}
			}
		
		}

	public void createMethodTable() {
		createMethodTable( methodTable );
		}
		
	public void setLocalOffset() {
		MethodDeclNode methodDeclNode = ( MethodDeclNode) decl.declNode();
		int weight = Math.max( methodDeclNode.body().weight() - Node.MAXREG, 0 );
		localOffset = new Offset( 3 + weight );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_LOCAL:
					decl.setOffset( localOffset );
				}
			}
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_VARPARAM:
				case Decl.SECT_VALUEPARAM:
					decl.setOffset( localOffset );
				}
			}
		}

	public void genLocalDecls() {
		Node.enter( "local" );

		Node.publicLabelDefn( "locals" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_LOCAL:
					decl.allocateSpace();
				}
			}
		Node.publicLabelDefn( "params" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_VARPARAM:
				case Decl.SECT_VALUEPARAM:
					decl.allocateSpace();
				}
			}
		Node.publicLabelDefn( "max" );
		Node.exit();
		}
		
	public void setMemberOffset() {
		if ( extendedEnv == null ) {
			instanceFieldOffset = new Offset( 0 );
			instanceCodeOffset = new Offset( 0 );
			}
		else {
			instanceFieldOffset = extendedEnv.instanceFieldOffset.copy();
			instanceCodeOffset = extendedEnv.instanceCodeOffset.copy();
			}
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_CLASS:
					decl.setOffset( classOffset );
					break;
				case Decl.SECT_STATIC:
				case Decl.SECT_GLOBAL:
					if ( decl.type() instanceof MethodType )
						decl.setOffset( staticCodeOffset );
					else
						decl.setOffset( staticFieldOffset );
					break;
				case Decl.SECT_INSTANCE:
					if ( decl.type() instanceof MethodType )
						decl.setOffset( instanceCodeOffset );
					else
						decl.setOffset( instanceFieldOffset );
					break;
				}
			}
		}

	public void genObjectFieldDecls() {
		if ( extendedEnv == null )
			Node.enterBlock( "field", null, null );
		else
			Node.enterBlock( "field", extendedEnv.absoluteName() + ".object.field", extendedEnv.absoluteName() + ".object.field" );
		Node.enter( "local" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_INSTANCE:
					if ( ! ( decl.type() instanceof MethodType ) )
						decl.allocateSpace();
					break;
				}
			}
		Node.publicLabelDefn( "max" );
		Node.exit();
		Node.exit();
		}

	public void genObjectCodeDecls() {
		if ( extendedEnv == null )
			Node.enterBlock( "method", null, null );
		else
			Node.enterBlock( "method", extendedEnv.absoluteName() + ".object.method", extendedEnv.absoluteName() + ".object.method" );
		Node.enter( "local" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_INSTANCE:
					if ( decl.type() instanceof MethodType && decl.overRiddenDecl() == null )
						decl.allocateSpace();
					break;
				}
			}
		Node.publicLabelDefn( "max" );
		Node.exit();
		Node.exit();
		}
		
	public void genObjectDecls() {
		Node.enterBlock( "object", null, null );
		genObjectFieldDecls();
		genObjectCodeDecls();
		Node.exit();
		}
		
	public void genInstanceCodeDecls() {
		if ( extendedEnv == null )
			Node.enterBlock( "method", null, null );
		else
			Node.enterBlock( "method", null, extendedEnv.absoluteName() + ".instance.method" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_INSTANCE:
					if ( decl.type() instanceof MethodType ) {
						decl.declNode().genDeclCode();
						}
					break;
				}
			}
		Node.exit();
		}
		
	public void genInstanceDecls() {
		Node.enterBlock( "instance", null, null );
		genInstanceCodeDecls();
		Node.exit();
		}
		
	public void genStaticFieldDecls() {
		Node.enterBlock( "field", null, null );
		Node.enter( "data" );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_GLOBAL:
				case Decl.SECT_STATIC:
					if ( ! ( decl.type() instanceof MethodType ) )
						decl.allocateSpace();
					break;
				}
			}
		Node.exit();
		Node.exit();
		}

	public void genStaticCodeDecls() {
		Node.enterBlock( "method", null, null );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_GLOBAL:
				case Decl.SECT_STATIC:
					if ( decl.type() instanceof MethodType ) {
						decl.declNode().genDeclCode();
						}
					break;
				}
			}
		Node.exit();
		}
		
	public void genStaticDecls() {
		Node.enterBlock( "static", null, null );
		genStaticFieldDecls();
		genStaticCodeDecls();
		Node.exit();
		}

	public void genClassDecls() {
		Node.enterBlock( "subClass", null, null );
		for ( int i = 0; i < localDecls.size(); i++ ) {
			Decl decl = localDecls.elementAt( i );
			switch ( decl.section() ) {
				case Decl.SECT_CLASS:
					decl.declNode().genDeclCode();
					break;
				}
			}
		Node.exit();
		}
		
	public void genMethodTable() {
		Node.enterBlock( "methodTable", null, null );
		Node.enter( "const" );
		Node.publicLabelDefn( "enter" );
		for ( int i = 0; i < methodTable.size(); i++ ) {
			Decl decl = methodTable.elementAt( i );
			Node.quad( decl.env().absoluteName() + ".instance.method." + decl.declName() + ".enter" );
			}
		Node.exit();
		Node.exit();
		}
	
	}
