package grammar;

import java.io.*;
import java_cup.runtime.*;

%%

%public
%type		Symbol
%char

%{
	private int lineNumber = 1;
	public int lineNumber() { return lineNumber; }
	
	public Symbol token( int tokenType ) {
		//	Print.error.println( "Obtain token " + sym.terminal_name( tokenType ) + " \"" + yytext() + "\"" );
		return new Symbol( tokenType, yychar, yychar + yytext().length(), yytext() );
		}
%}

%init{
	yybegin( NORMAL );
%init}

intconst	=	[0-9]+
exponent	=	[Ee][\+\-]?[0-9]+
float1		=	[0-9]+\.[0-9]+
float2		=	[0-9]+{exponent}
float3		=	[0-9]+\.[0-9]+{exponent}
octDigit	=	[0-7]
hexDigit	=	[0-9a-fA-F]
escchar		=	\\([ntbrfva\\\'\"\?]|{octDigit}+|[xX]{hexDigit}+)
schar		=	[^\'\"\\\r\n]|{escchar}
charconst	=	\'{schar}\'
stringconst	=	\"{schar}*\"
ident		=	[A-Za-z][A-Za-z0-9]*
space		=	[\ \t]
newline		=	\r|\n|\r\n
%state NORMAL COMMENT1 COMMENT2 LEXERROR

%%
<NORMAL> {
	{newline}	{ lineNumber++; }
	{space}		{ }
	
	print		{ return token( sym.PRINT ); }
	println		{ return token( sym.PRINTLN ); }
	if			{ return token( sym.IF ); }
	then		{ return token( sym.THEN ); }
	else		{ return token( sym.ELSE ); }
	elif		{ return token( sym.ELIF ); }
	var			{ return token( sym.VAR ); }
	new			{ return token( sym.NEW ); }
	begin		{ return token( sym.BEGIN ); }
	end			{ return token( sym.END ); }
	while		{ return token( sym.WHILE ); }
	do			{ return token( sym.DO ); }
	for			{ return token( sym.FOR ); }
	return		{ return token( sym.RETURN ); }
	size		{ return token( sym.SIZE ); }
	null		{ return token( sym.NULL ); }
	int			{ return token( sym.INT ); }
	char		{ return token( sym.CHAR ); }
	real		{ return token( sym.REAL ); }
	bool		{ return token( sym.BOOL ); }
	void		{ return token( sym.VOID ); }
	class		{ return token( sym.CLASS ); }
	static		{ return token( sym.STATIC ); }
	instance	{ return token( sym.INSTANCE ); }
	extends		{ return token( sym.EXTENDS ); }
	this		{ return token( sym.THIS ); }
	
	";"			{ return token( sym.SEMICOLON ); }
	","			{ return token( sym.COMMA ); }
	".."		{ return token( sym.DOTDOT ); }
	"="			{ return token( sym.ASSIGN ); }
	
	or			{ return token( sym.OR ); }
	and			{ return token( sym.AND ); }
	not			{ return token( sym.NOT ); }
	"<"			{ return token( sym.LT ); }
	">"			{ return token( sym.GT ); }
	"<="		{ return token( sym.LE ); }
	">="		{ return token( sym.GE ); }
	"=="		{ return token( sym.EQ ); }
	"<>"		{ return token( sym.NE ); }
	"+"			{ return token( sym.PLUS ); }
	"-"			{ return token( sym.MINUS ); }
	"*"			{ return token( sym.TIMES ); }
	"/"			{ return token( sym.DIVIDE ); }
	"%"			{ return token( sym.MOD ); }
	"++"		{ return token( sym.INC ); }
	"--"		{ return token( sym.DEC ); }
	"("			{ return token( sym.LEFT ); }
	")"			{ return token( sym.RIGHT ); }
	"["			{ return token( sym.LEFTSQ ); }
	"]"			{ return token( sym.RIGHTSQ ); }
	"{"			{ return token( sym.LEFTCURLY ); }
	"}"			{ return token( sym.RIGHTCURLY ); }
	"."			{ return token( sym.DOT ); }
	
	{intconst}	{ return token( sym.INTVALUE ); }
	{charconst}	{ return token( sym.CHARVALUE ); }
	{stringconst}	{ return token( sym.STRINGVALUE ); }
	{float1}	{ return token( sym.REALVALUE ); }
	{float2}	{ return token( sym.REALVALUE ); }
	{float3}	{ return token( sym.REALVALUE ); }
	{ident}		{ return token( sym.IDENT ); }
	"//"		{ yybegin( COMMENT1 ); }
	"/*"		{ yybegin( COMMENT2 ); }
	.			{
				yybegin( LEXERROR ); 
				return token( sym.error ); 
				}
	}
	
<LEXERROR> {
	";"			{
				yybegin( NORMAL ); 
				return token( sym.SEMICOLON ); 
				}
	{newline}	{
				lineNumber++;
				yybegin( NORMAL ); 
				}
	.			{ }
	}

<COMMENT1> {
	{newline}	{ 
				lineNumber++;
				yybegin( NORMAL ); 
				}
	.			{ }
	}
	
<COMMENT2> {
	"*/"		{ yybegin( NORMAL ); }
	{newline}	{ 
				lineNumber++;
				}
	.			{ }
	}

<<EOF>>			{ return token( sym.EOF ); }

